
# dsp2vmake.py

import sys
import os
import string
import re
from cStringIO import StringIO
from UserDict import UserDict

import optlib

#
# import specialCases.py (should be in $SRC)
#
src = None
try:
    src = os.environ ['SRC']
    sys.path.insert (0, src)
except KeyError:
    pass

import specialCases

if src:
    del sys.path [0]

#
# import imkdepLib
#
try:
    import vmake.imkdepLib
    dependencySection = vmake.imkdepLib.DepUpdater.dependencySection
except ImportError:
    dependencySection = """#
# include dependencies (do not edit)
#
"""

#
# import imkdep
#
imkdep = None
if os.environ.get ('SRC') and os.environ.get ('DESC'):
    try:
        import imkdep
    except ImportError:
        pass

true = 'true'
false = None

def fatalError (msg):
    sys.stderr.write ('dsp2vmake.py: %s\n' % msg)

def trace (fmt, *args):
    if args:
        line = fmt % args
    else:
        line = '%s' % fmt
    sys.stdout.write (line + '\n')

def normpath (fname):
    return os.path.normpath (os.path.normcase (fname))

def extension (fname):
    return os.path.splitext (fname) [1]


def smartOpen (fname, mode):
    dirname = os.path.dirname (fname)
    if not os.path.exists (dirname):
        os.makedirs (dirname)
    return open (fname, mode)

_includeExtensions = ['.h', '.hpp']
_allowExtensions = ['.c', '.cpp', '.cxx', '.tli', '.h', '.tlh',
    '.inl', '.rc', '.def', '.idl']
_denyExtensions = []

def allowFileByExtension (fname):
    extension = os.path.splitext (fname) [1]
    if len (_allowExtensions) > 0:
        return extension in _allowExtensions
    if len (_denyExtensions) > 0:
        return extension not in _denyExtensions
    return true

class Dsp2Vmake:
    rcRE = re.compile ('.*[.]rc$')
    defRE = re.compile ('.*[.]def$')
    def __init__ (self, dollarSrc, dollarDesc, knownLibs, options):
        self.dollarSrc = normpath (dollarSrc)
        self.srcLen = len (self.dollarSrc) + 1
        self.dollarDesc = normpath (dollarDesc)
        self.descLen = len (self.dollarDesc) + 1
        self.knownLibs = knownLibs
        self.options = options
        self.layer = LayerDict ()
        self.optionSampler = OptionSampler ()
        self.specialTargetHandling = [
            (self.rcRE, self.win32OnlyWrite),
            (self.defRE, self.win32OnlyWrite),
        ]
        self.nameTranslations = [
            (re.compile ('(.*)Ispc[.]cpp$'), r'\1.idl'),
        ]
        self.additionalLayer = specialCases.getAdditionalLayer (options.inDEV)
        self.specialCases = specialCases.cases
        self.specialLibraries = self.compileSpecialLibraries (specialCases.specialLibraries)
        # self.externalLibraries = specialCases.externalLibraries

    def compileSpecialLibraries(self, specialLibraries):
        result = []
        for pattern, template in specialLibraries:
            result.append ((re.compile (pattern), template))
        return result


    def getTargetKinds (self, args):
        if args:
            dspFiles = []
            for arg in args:
                ext = os.path.splitext (arg) [1]
                if ext == '.dsw':
                    parser = WorkspaceParser (arg)
                    for dsp, dependencies in parser.getDependencies ():
                        dspFiles.append (dsp)
                else:
                    dspFiles.append (os.path.abspath (arg))
        else:
            dspFiles = self.getAllDSP (self.dollarSrc)
        rex = DspFile.reTargetType
        self.knownLibs = {}
        for dsp in dspFiles:
            data = open (dsp, 'r').read ()
            match = rex.search (data)
            if match:
                ext = self.targetTypeMap [match.group (1)]
                most, dspExt = os.path.splitext (dsp)
                base = os.path.basename (most)
                relative = '::' + most [self.srcLen:] + ext
                relative = string.replace (relative,  '\\', '/')
                key = string.lower (base)
                self.knownLibs [key] = relative

    def translate (self, inDEV):
        """translates all .dsp in inDEV to equivalent
            vmake descriptions in outSys/desc
        """
        if inDEV [-1] in "/\\":
            inDEV = inDEV [:-1]
        dspFiles = self.getAllDSP (inDEV)
        for absolute in dspFiles:
            self.translateOneFile (absolute)

    def getAllDSP (self, inDEV):
        """gets the names of all .dsp files in inDEV

        returns: a list containing both the names relative
            to inDEV and the absolute names.
        """
        result = []
        def filterDSP (list, dirname, names):
            for name in names:
                if name [-4:] == '.dsp':
                    fullpath = os.path.join (dirname, name)
                    list.append (fullpath)
        os.path.walk (inDEV, filterDSP, result)
        result = filter (self.isValidPattern, result)
        return result

    def translateWorkspace (self, workspaceFile):
        parser = WorkspaceParser (workspaceFile)
        for dsp, dependencies in parser.getDependencies ():
            self.translateOneFile (dsp, dependencies)

    def isValidPattern (self, filename):
        for pattern in self.options.ignorePatterns:
            if string.find (string.lower (filename), pattern) != -1:
                trace ('ignoring %s because of %s', filename, pattern)
                return None
        return 1

    targetTypeMap = {
        'Console Application': '.lnk',
        'Application': '.lnk',
        'Dynamic-Link Library': '.dld',
        'Generic Project': '.mac',
        'Static Library': '.lib',
    }
    def translateOneFile (self, absolute, dependencies = []):
        dirname = os.path.dirname (absolute)
        #
        # parse input file
        #
        if self.options.verbose:
            trace (absolute)
        try:
            dspFile = DspFile (absolute, self.optionSampler, self.options.smartVariants)
        except DspParseError, err:
            sys.stderr.write (str (err) + '\n')
            return
        #
        # create output file
        #
        if not dspFile.targetType:
            #if self.options.verbose:
            #    print 'skipping'
            return
        targetExtension = self.targetTypeMap [dspFile.targetType]
        relativeTarget = absolute [self.srcLen:-4] + targetExtension
        if targetExtension in ['.lib', '.dld']:
            shortAlias = string.lower (split3 (relativeTarget) [1])
            self.knownLibs [shortAlias] = '::' + string.replace (relativeTarget, '\\', '/')
        outputFile = os.path.join (self.dollarDesc, relativeTarget)
        stream = smartOpen (outputFile, 'wt')
        stream.write ('# ' + string.replace (relativeTarget, '\\', '/') + '\n\n')
        #
        # write options
        #
        windowsLibs, binaryLibs, madeLibs = self.getReferencedLibs (dspFile)
        if len (windowsLibs) > 0:
            stream.write ('&if $OS == WIN32\n')
            libsWritten = []
            for libname, variant in windowsLibs:
                # self.writeVariant (stream, '-lib=' + libname, variant)
                # make it a bit simpler, it's variant independent anyway
                if libname not in libsWritten:
                    self.writeVariant (stream, '-lib=' + libname, stdVariant)
                    libsWritten.append (libname)
            stream.write ('&endif\n')
        if len (binaryLibs) > 0:
            for replacement, variant in binaryLibs:
                self.writeVariant (stream, replacement, variant)
        #
        # write targets
        #
        localFiles = StringIO ()
        hasLocalFiles = None
        localFiles.write ('\n? defaultlayer : ' +
            string.replace (dirname [self.srcLen:], '\\', '/') + '\n')
        otherFiles = StringIO ()
        lastDir = ''
        for source in dspFile.noCompileSources ():
            sourceFilename = self.translateFilename (source.name)
            fullpath = os.path.normpath (os.path.join (dirname, sourceFilename))
            dirpart, basepart = self.splitpath (fullpath)
            self.layer.addEntry (dirpart, basepart, source, dspFile.filename)
            ext = extension (sourceFilename)
            if ext in _includeExtensions:
                continue
            if not allowFileByExtension (sourceFilename):
                continue
            source.removeIgnoredVariants (self.options.ignoreVariant)
            source.restrictToOnlyVariants (self.options.onlyVariant)
            emit, variants = self.getVariants (source)
            if not emit:
                continue
            addOn = self.moduleAddOn (ext)
            dirpart, basepart = os.path.split (fullpath)
            if dirpart == dirname:
                self.writeTarget (localFiles,
                    source.currentConfig.filenameVariant (basepart),
                    addOn, variants)
                hasLocalFiles = 1
            else:
                srcPath = string.replace (fullpath [self.srcLen:], '\\', '/')
                srcDir = os.path.dirname (srcPath)
                if srcDir != lastDir:
                    otherFiles.write ('\n')
                    lastDir = srcDir
                self.writeTarget (otherFiles,
                    ':' + source.currentConfig.filenameVariant (srcPath),
                    addOn, variants)
        if hasLocalFiles:
            stream.write (localFiles.getvalue ())
            stream.write ('\n')
        stream.write (otherFiles.getvalue ())
        if len (madeLibs) > 0:
            stream.write ('\n')
            for libname, variant in madeLibs:
                self.writeVariant (stream, self.findKnownLib (libname), variant)
        if dependencies:
            stream.write ('\n')
        for depfile in dependencies:
            relative = depfile [self.srcLen:]
            relative = string.replace (relative, '\\', '/')
            relative = os.path.splitext (relative) [0]
            stream.write ('::' + relative + ' nobind\n')
        self.addFileSpecifics (stream, relativeTarget)
        stream.close ()

    def getVariants (self, source):
        variants = []
        excludeCount = 0
        for config in source.getConfigs ():
            exclude = config.properties.get ('Exclude_From_Build')
            if exclude:
                excludeCount = excludeCount + 1
            variants.append ((config.variantNames, not exclude))
        if excludeCount == 0:
            result = (1, None)
        elif excludeCount == len (variants):
            result = (0, None)
        else:
            result = (1, variants)
        return result

    def findKnownLib (self, libname):
        key = os.path.basename (libname)
        key = string.lower (os.path.splitext (key) [0])
        knownLibs = self.knownLibs
        if knownLibs.has_key (key):
            return knownLibs [key]
        if key [-1] == 'd':
            key = key [:-1]   # try as debug lib
            if knownLibs.has_key (key):
                return knownLibs [key]
        return libname

    def translateFilename (self, sourceName):
        if os.sep != '\\':
            sourceName = string.replace (sourceName, '\\',  os.sep)
        for rex, replacement in self.nameTranslations:
            if rex.match (sourceName):
                return rex.sub (replacement, sourceName)
        return sourceName

    def writeTarget (self, stream, filename, addOn, variants):
        for rex, function in self.specialTargetHandling:
            if rex.match (filename):
                function (stream, filename, addOn, variants)
                break
        else:
            self.normalTargetWrite (stream, filename, addOn, variants)

    def normalTargetWrite (self, stream, filename, addOn, variants):
        line = filename + addOn + '\n'
        if not variants:
            stream.write (line)
        else:
            variants.sort ()
            variants.reverse ()
            variantsWritten = []
            intro = '&if '
            for variant, include in variants:
                variant = filter (None, variant)
                if variant in variantsWritten:
                    continue
                variantsWritten.append (variant)
                if variant:
                    list = map (lambda var: 'variant (' + var + ')', variant)
                    str = string.join (list, ' && ')
                    stream.write ('%s %s\n' % (intro, str))
                else:
                    stream.write ('&else\n')
                if include:
                    stream.write (line)
                else:
                    stream.write ('# Exclude_From_Build\n')
                intro = '&elif '
            stream.write ('&endif\n')

    def win32OnlyWrite (self, stream, filename, addOn, variants):
        stream.write ('&if $OS == WIN32\n')
        self.normalTargetWrite (stream,  filename,  addOn,  variants)
        stream.write ('&endif\n')

    def getReferencedLibs (self, dspFile):
        windowsLibs = []
        binaryLibs = []
        madeLibs = []
        moduleOptions = dspFile.moduleOptions
        for config in moduleOptions.configurations.values ():
            for option, value in config.linkOptions:
                if option == '-lib':
                    replacement = self.handleSpecialLibs (value)
                    if replacement:
                        targetList = binaryLibs
                        value = replacement
                    elif value [0] == '.':
                        targetList = madeLibs
                    else:
                        targetList = windowsLibs
                    if os.sep != '\\':
                        value = string.replace (value,  '\\',  os.sep)
                    entry = (value, config.variantNames)
                    if entry not in targetList:
                        targetList.append (entry)
        #windowsLibs.sort ()
        #binaryLibs.sort ()
        #madeLibs.sort ()
        return windowsLibs, binaryLibs, madeLibs

    def handleSpecialLibs (self, libname):
        for rex, template in self.specialLibraries:
            match = rex.match (libname)
            if match:
                replacement = reExpand (template, rex, match)
                return replacement
        return None

    def writeVariant (self, stream, line, variantNames):
        if variantNames == stdVariant:
            stream.write (line + '\n')
        else:
            variantNames = filter (None, variantNames)
            list = map (lambda var: 'variant (' + var + ')', variantNames)
            str = string.join (list, ' && ')
            stream.write ('&if ' + str + '\n')
            stream.write (line + '\n')
            stream.write ('&endif\n')

    def moduleAddOn (self, extension):
        if extension == '.def':
            result = ' definition'
        elif extension == '.rc':
            result = ' remake'
        else:
            result = ''
        return result

    def splitpath (self, absolute):
        relative = absolute [self.srcLen:]
        return os.path.split (relative)

    def writeComDescriptions (self):
        layer = self.layer.items ()
        for path in self.additionalLayer:
            layer.append ((path, []))
        layer.sort ()
        for path, entries in layer:
            if self.options.unifiedSys:
                fname = os.path.join (path, os.path.basename (path) + '.com')
            else:
                fname = path + '.com'
            outfile = os.path.join (self.dollarDesc, fname)
            if self.options.verbose:
                trace (outfile)
            stream = smartOpen (outfile, 'wt')
            stream.write ('# ' + string.replace (fname, '\\', '/') + '\n')
            for glob in self.options.globals:
                stream.write (glob + '\n')
            if len (entries) > 0:
                entries = self.detectConflictingComEntries (entries)
                for entry, configuredItem, dspFile in entries:
                    configuredItem.removeIgnoredVariants (self.options.ignoreVariant)
                # self.dumpDescriptionEntries (entries)
                starDescription = self.extractCommonOptions (entries, path)
                self.writeDescriptionVariants ('*', starDescription, None, stream)
                entries.sort ()
                for entry, configuredItem, dspFile in entries:
                    configuredItem.normalizeOptions ()
                for entry, configuredItem, dspFile in entries:
                    self.writeDescriptionVariants (entry, configuredItem, dspFile, stream)
            #
            # add platform specific rules
            #
            self.addFileSpecifics (stream, fname)
            stream.write (dependencySection)
            stream.close ()
            #
            # handle include dependencies
            #
            if imkdep and not self.options.nodep:
                args = ['-noinames', '-unifiedSys', outfile]
                #if self.options.verbose:
                #    args.insert (0, '-v')
                optlib.simulateMain (imkdep,  args)

    def addFileSpecifics (self, stream, relative):
        relative = string.replace (relative,  '\\',  '/')
        for case in self.specialCases:
            if relative == case.filename:
                case.apply (stream, self.options)
        fullpath = os.path.join (self.dollarDesc, relative + '.plus')
        if os.path.exists (fullpath):
            stream.write ("\n# included from %s.plus\n" % relative)
            data = open (fullpath, 'r').read ()
            stream.write (data)
            stream.write ("# end include\n")

    def dumpDescriptionEntries (self, entries):
        for entry, configuredItem in entries:
            trace ('=== %s', configuredItem.name)
            for config in configuredItem.configurations.values ():
                trace ('   %s %s', config.variantNames, config.options)

    def detectConflictingComEntries (self, entries):
        entries.sort ()
        result = []
        lastEntry, lastItem, lastDsp = entries [0]
        result.append ((lastEntry, lastItem, lastDsp))
        for entry, configuredItem, dspFile in entries [1:]:
            if entry == lastEntry:
                for config in configuredItem.getConfigs ():
                    last = lastItem.configurations.get (config.variantNames)
                    if not last:
                        # variant not in last entry, add to list
                        lastItem.configurations [config.variantNames] = config
                    else:
                        remLast, common, remConfig = setSplit (last.options, config.options)
                        if remLast or remConfig:
                            if self.options.warn:
                                trace ('Conflict: %s variant (%s) common: %s',  entry, config.variantNames, common)
                                trace ('--- %s: %s',  lastDsp, remLast)
                                trace ('--- %s: %s', dspFile, remConfig)
            else:
                lastEntry, lastItem, lastDsp = lastTuple = entry, configuredItem, dspFile
                result.append (lastTuple)
        return result

    def extractCommonOptions (self, entries, path):
        allVariantNames = {}
        for entry, configuredItem, dspFile in entries:
            for variantName in configuredItem.getVariantNames ():
                allVariantNames [variantName] = 1
        starDescription = ConfiguredItem ('*')
        del starDescription.configurations [stdVariant]
        allVariantNames = allVariantNames.keys ()
        allVariantNames.sort ()
        if path in ['atp\\atplib', 'atp/atplib']:
            allVariantNames.reverse ()
        for variantName in allVariantNames:
            variantConfigs = self.filterVariantOptions (entries, variantName)
            commons = variantConfigs.extractCommonOptions (variantName)
            for entry in commons:
                starDescription.configurations [entry.variantNames] = entry
        if not starDescription.configurations.has_key (stdVariant):
            # make sure that an empty common definition gets written
            stdConfig = Configuration (stdVariant)
            stdConfig.isLocked = 1
            starDescription.configurations [stdVariant] = stdConfig
        starDescription.normalizeOptions ()
        return starDescription

    def filterVariantOptions (self, entries, variantName):
        list = []
        for entry, configuredItem, dspFile in entries:
            list.extend (configuredItem.filter (variantName, true))
        result = VariantSet (list)
        return result


    def dumpEntries (self, entries):
        for entry, configuredItem, dspFile in entries:
            trace (entry)
            configuredItem.dump ()

    def findOptionCandidates (self, entries, configId):
        for entry, configuredItem, dspFile in entries:
            try:
                config = configuredItem.configurations [configId]
                return config.options [:]
            except KeyError:
                pass

    def isCommonOption (self, entries, configId, optdef):
        result = true
        for entry, configuredItem, dspFile in entries:
            try:
                config = configuredItem.configurations [configId]
                if optdef not in config.options:
                    return false
            except KeyError:
                pass
        return result

    def removeCommonOption (self, entries, configId, optdef):
        result = true
        for entry, configuredItem, dspFile in entries:
            try:
                config = configuredItem.configurations [configId]
                config.options.remove (optdef)
            except KeyError:
                pass
        return result

    def writeDescriptionVariants (self, entry, configuredItem, dspFile, stream):
        if dspFile and self.options.debug:
            stream.write ('# from ' + dspFile + '\n')
        count = 0
        for config in configuredItem.extraConfigs ():
            line = self.sourceDescription (entry, config)
            if line:
                stream.write (line + '\n')
                count = count + 1
        if (count == 0) and self.options.debug:
            stream.write ('# %s\n' % entry)

    def sourceDescription (self, fname, configuration, unconditional = false):
        #if not configuration.options and not unconditional \
        #        and configuration.variantNames != stdVariant:
        #    return '# ' + fname + ' ' + `configuration.variantNames` + '\n'
        #    return None
        result = fname
        if configuration.variantNames != stdVariant:
            varstring = string.join (filter (None,  configuration.variantNames),  ',')
            result = result + ':variant(' + varstring + ')'
        for option, value in configuration.options:
            result = result + ' -' + option [1:]
            if value:
                result = result + unquote (value)
        return result


class DspFile:
    """Parser for MS Dev project files
    """
    allowSource = true
    optionsToIgnore = specialCases.optionsToIgnore
    shortOptionsToIgnore = specialCases.shortOptionsToIgnore
    definesToIgnore = specialCases.definesToIgnore

    def __init__ (self, fname, optionSampler, smartVariants = None):
        self.filename = fname
        self.targetType = None
        self.moduleOptions = ConfiguredItem ('')
        self.currentSource = self.moduleOptions
        self.sources = []
        matchList = []
        for kind in ['TargetType', 'Source', 'EndSource', 'Add', 'Prop',
                'If', 'EndIf', 'SpecialBuild', 'EndSpecialBuild']:
            matchList.append ((getattr (self, 're' + kind), getattr (self, 'match' + kind)))
        self.matcher = RegexEvents (matchList)
        self.optionSampler = optionSampler
        self.smartVariants = smartVariants
        self.inConditionalConfig = None
        #
        self.parseFile (fname)

    def noCompileSources (self):
        result = []
        for source in self.sources:
            appendSource = true
            filetype = os.path.splitext (source.name) [1]
            if filetype in ['.h', '.hpp']:
                appendSource = false
            elif source.name == '$(InputPath)':
                appendSource = false
            if appendSource:
                result.append (source)
        return result


    def parseFile (self, fname):
        lines = open (fname, 'rt').readlines ()
        self.matcher.parseLines (lines)
        for item in self.sources:
            item.mergeCommonEntries (self.moduleOptions)

    reTargetType = re.compile ('# TARGTYPE "Win32 [(]x86[)] ([^"]+)" 0x([0-9]+)')
    def matchTargetType (self, match):
        self.targetType = match.group (1)

    reSource = re.compile ('SOURCE=(.+)')
    def matchSource (self, match):
        if self.allowSource:
            filename = match.group (1)
            newItem = ConfiguredItem (filename)
            self.sources.append (newItem)
            self.currentSource = newItem

    reEndSource = re.compile ('# End Source File')
    def matchEndSource (self, match):
        self.currentSource = self.moduleOptions

    reSpecialBuild = re.compile ('# Begin Special Build Tool')
    def matchSpecialBuild (self, match):
        self.allowSource = false

    reEndSpecialBuild = re.compile ('# End Special Build Tool')
    def matchEndSpecialBuild (self, match):
        self.allowSource = true

    reAdd = re.compile ('# ADD (.*)')
    def matchAdd (self, match):
        line = match.group (1)
        line = string.replace (line, 'Program Files', 'Program_Files')
        line = string.replace (line, 'Microsoft Visual Studio', 'Microsoft_Visual_Studio')
        words = string.split (line)
        if words [0] == 'BASE':
            return   # BASE options are obsolete
            del words [0]
        tool = words [0]
        del words [0]
        parse = {'CPP': self.parseCPPOptions, 'LINK32': self.parseLINK32Options}\
            .get (tool, self.parseAnyOptions)
        for option, value in parse (words):
            self.optionSampler.addOption (tool, option, value)
            if tool == 'CPP':
                self.currentSource.addOption (option, value)
            elif tool == 'LINK32':
                self.currentSource.addLinkOption (option, value)

    variantRE = re.compile ('"?VMAKEVARIANT_([a-zA-Z0-9_]+)"?')
    def parseCPPOptions (self, words):
        result = []
        if words [0] == 'BASE':
            return   # BASE options are obsolete
            index = 1
        else:
            index = 0
        words.append ('/') # makes looking for value easier
        while index < len (words) - 1:
            option = words [index]
            value = words [index + 1]
            if value [0] == '/':
                value = None
            else:
                index = index + 1
            storeValue = None
            if (option in self.optionsToIgnore
                    or option [:2] in self.shortOptionsToIgnore):
                # common options are set through toolvars.pl or variants
                pass
            elif option [:2] == '/I':
                if value [:2] == '".':
                    # relative paths are ignored, replaced be /I %WRK%/incl
                    pass
                elif value == '"$(LC_INC)"':
                    # don't know how to make this common
                    pass
                elif string.find (value, 'VC98') != -1:
                    # this one isn't necessary anyway
                    pass
                else:
                    storeValue = true
            elif option [:2] == '/D':
                if len (option) > 2:
                    value = option [2:]
                    option = '/D'
                if value in self.definesToIgnore:
                    pass
                else:
                    storeValue = true
                    if value:
                        match = self.variantRE.match (value)
                        if match:
                            variant = match.group (1)
                            self.currentSource.switchToCopy (variant)
                            self.currentSource.addProperty ('VMAKEVARIANT', (variant,))
                        elif string.find (value, 'LATEST_CHANGE') != -1:
                            value = 'LATEST_CHANGE=$LATEST_CHANGE'
                            # much too common
                            storeValue = false
                        elif string.find (value, 'BRANCH_INFO') != -1:
                            value = 'BRANCH_INFO=$BRANCH_INFO'
            elif option [:2] == '/Y':
                # currently no support for precompiled header
                pass
            else:
                storeValue = true
            if storeValue:
                result.append ((option, value))
            index = index + 1
        return result

    def parseLINK32Options (self, words):
        result = []
        for word in words:
            if word [0] == '/':
                colonPos = string.find (word, ':')
                if colonPos == -1:
                    option = word
                    value = None
                else:
                    option = word [:colonPos]
                    value = unquote (word [colonPos + 1:])
            else:
                option = '-lib'
                value = word
            result.append ((option, value))
        return result


    def parseAnyOptions (self, words):
        result = []
        index = 1
        words.append ('/') # makes looking for value easier
        while index < len (words) - 1:
            option = words [index]
            value = words [index + 1]
            if value [0] == '/':
                value = None
            else:
                index = index + 1
            index = index + 1
            result.append ((option, value))
        return result

    reProp = re.compile ('# PROP (.*)')
    def matchProp (self, match):
        words = string.split (match.group (1))
        if words [0] == 'BASE':
            return   # BASE options are obsolete
            del words [0]
        assert len (words) == 2, 'more than 2 words in # PROP ' + `words`
        self.currentSource.addProperty (words [0], words [1])

    reIf = re.compile ('!(ELSE)?IF  "[$][(]CFG[)]" == "([^"]+)"')
    def matchIf (self, match):
        configName = match.group (2)
        shortName = string.split (configName) [-1]
        configId = (shortName, '')
        if self.smartVariants:
            for special in ['Release', 'Debug']:
                splen = len (special)
                if shortName [:splen] == special:
                    part1 = shortName [:splen]
                    part2 = shortName [splen:]
                    configId = (part1, part2)
                    break
        self.currentSource.setCurrentConfiguration (configId)
        self.inConditionalConfig = 1

    reEndIf = re.compile ('!ENDIF')
    def matchEndIf (self, match):
        if self.inConditionalConfig:
            self.currentSource.setStdConfiguration ()
            self.inConditionalConfig = None

class DspParseError:
    def __init__ (self, msg, fname):
        self.msg = msg
        self.fname = fname

    def __str__ (self):
        return '.dsp Parse error: ' + self.msg + ' in ' + self.fname

class RegexEvents:
    def __init__ (self, reList):
        self.reList = reList

    def parseLines (self, lines):
        reList = self.reList
        for line in lines:
            for rex, action in reList:
                match = rex.match (line)
                if match:
                    action (match)

    def parseFile (self, fname):
        self.parseLines (open (fname, 'rt').readlines ())

stdVariant = ('',)

class ConfiguredItem:
    def __init__ (self, name):
        if name:
            name = unquote (name)
        self.name = name
        self.configurations = {}
        self.setStdConfiguration ()

    def __str__ (self):
        return ('[' + self.name + ' at '
            + `self.currentConfig.variantNames` + ' '
            + `self.configurations.keys ()` + ']')


    def setCurrentConfiguration (self, configId, extended = false):
        try:
            self.currentConfig = self.configurations [configId]
        except KeyError:
            self.currentConfig = Configuration (configId, extended)
            configId = self.currentConfig.variantNames
            self.configurations [configId] = self.currentConfig
        return self.currentConfig

    def setStdConfiguration (self):
        self.setCurrentConfiguration (stdVariant)

    def switchToCopy (self, variantName):
        assert self.currentConfig.variantNames == stdVariant, 'Can\'t handle nested variants yet'
        # variantNames = (variantName,) + self.currentConfig.variantNames
        self.setCurrentConfiguration ((variantName,), true)

    def addOption (self, option, value = None):
        optdef = (option, value)
        currentOptions = self.currentConfig.options
        if optdef not in currentOptions:
            currentOptions.append (optdef)

    def addLinkOption (self, option, value = None):
        optdef = (option, value)
        currentOptions = self.currentConfig.linkOptions
        if optdef not in currentOptions:
            currentOptions.append (optdef)

    def addProperty (self, name, value):
        self.currentConfig.properties [name] = value

    def extraConfigs (self):
        result = []
        for config in self.configurations.values ():
            if config.variantNames != '':
                result.append (config)
        return result

    def getConfigs (self):
        return self.configurations.values ()

    def getVariantNames (self):
        result = []
        for config in self.configurations.values ():
            result.extend (list (config.variantNames))
        return result

    def normalizeOptions (self):
        self.clearEmptyConfigs ()
        set = VariantSet (self.configurations.values ())
        commons = set.extractCommonOptions ()
        self.clearEmptyConfigs ()
        for entry in commons:
            self.configurations [entry.variantNames] = entry

    def clearEmptyConfigs (self):
        for id, config in self.configurations.items ():
            if not config.options and not config.isLocked:
                del self.configurations [id]

    def mergeCommonEntries (self, otherConfiguredItem):
        defaultConfig = otherConfiguredItem.configurations [stdVariant]
        isEmpty = self.configurations.keys () == [stdVariant]
        #
        # merge global variants into self
        #
        for config in otherConfiguredItem.configurations.values ():
            try:
                localConfig = self.configurations [config.variantNames]
                self.mergeOneVariant (localConfig,  config)
            except KeyError:
                if isEmpty:
                    self.configurations [config.variantNames] = config.copy ()
        #
        # merge global variants into local variants
        #
        if not isEmpty and (len (self.configurations.keys ()) == 2):
            self.clearEmptyConfigs ()
            variantConfig = self.configurations.values () [0]
            if '' not in variantConfig.variantNames:
                # append global variants to variant configs
                for config in otherConfiguredItem.configurations.values ():
                    if config.variantNames == stdVariant:
                        continue
                    oldNames = variantConfig.variantNames
                    newNames = oldNames + config.variantNames
                    newConfig = self.setCurrentConfiguration (newNames)
                    newConfig.options = variantConfig.options + config.options
                    newConfig.properties.update (variantConfig.properties)
                    newConfig.properties.update (config.properties)
                del self.configurations [oldNames]
        #
        # merge defaults into local-only variants
        #
        for localConfig in self.configurations.values ():
            if otherConfiguredItem.configurations.has_key (localConfig.variantNames):
                pass # already handled by code above
            else:
                if '' not in localConfig.variantNames:
                    # extended variant, handled in next loop
                    continue
                self.mergeOneVariant (localConfig,  defaultConfig)
        #
        # merge local defaults into local variants
        #
        if self.configurations.has_key (stdVariant):
            localDefaults = self.configurations [stdVariant]
            for localConfig in self.configurations.values ():
                if localConfig.variantNames != stdVariant:
                    self.mergeOneVariant (localConfig,  localDefaults)
        #
        #
        #
        self.clearEmptyConfigs ()

    def mergeOneVariant (self, localConfig, config):
        for optdef in config.options:
            if optdef not in localConfig.options:
                localConfig.options.insert (0, optdef)
        for name, value in config.properties.items ():
            if not localConfig.properties.has_key (name):
                localConfig.properties [name] = value

    def removeDuplicateOptions (self):
        for config in self.configurations.values ():
            list = []
            for optdef in config.options:
                if optdef not in list:
                    list.append (optdef)
            config.options = list

    def extractCommonEntries (self):
        commonConfig = self.configurations [stdVariant]
        otherConfigs = self.extraConfigs ()
        if len (otherConfigs) == 0:
            return
        #
        # move common options to commonConfig
        #
        candidates = otherConfigs [0].options [:]
        for optdef in candidates:
            if self.isCommonOption (optdef):
                commonConfig.options.append (optdef)
                for config in otherConfigs:
                    config.options.remove (optdef)
        #
        # move common properties to commonConfig
        #
        for name, value in otherConfigs [0].properties.items ():
            if self.isCommonProperty (name, value):
                commonConfig.properties [name] = value
                for config in otherConfigs:
                    del config.properties [name]

    def isCommonProperty (self, name, value):
        result = true
        for config in self.configurations.values ():
            if config.variantNames == '':
                continue
            if config.properties.get (name) != value:
                return false
        return result

    def filter (self, variantName, returnList = false):
        return VariantSet (self.configurations.values ()).filter (variantName,  returnList)

    def removeIgnoredVariants (self, ignoreList):
        for variantNames in self.configurations.keys ():
            for name in ignoreList:
                if name in variantNames:
                    del self.configurations [variantNames]
                    continue

    def restrictToOnlyVariants (self, onlyList):
        if not onlyList:
            # empty list => generate all
            return
        for variantNames in self.configurations.keys ():
            for name in variantNames:
                if name and (name not in onlyList):
                    del self.configurations [variantNames]
                    break

    def dump (self, comment = ''):
        trace ('=== %s %s', self.name, comment)
        items = self.configurations.items ()
        items.sort ()
        for name, config in items:
            trace ('   %s %s %s', config.variantNames, config.options, sortedDict (config.properties))

class Configuration:
    isLocked = None
    def __init__ (self, variantNames, extended = false):
        assert type (variantNames) == type (())
        if ('' not in variantNames) and not extended:
            variantNames = variantNames + ('',)
        self.variantNames = variantNames
        self.options = []
        self.properties = {}
        self.linkOptions = []

    def __repr__ (self):
        return '<Configuration %s\n\t%s>' % (
            self.variantNames, self.options)

    def copy (self):
        result = Configuration (self.variantNames)
        result.options = self.options [:]
        result.properties = self.properties.copy ()
        return result

    def filenameVariant (self, filename):
        if self.properties.has_key ('VMAKEVARIANT'):
            name, ext = os.path.splitext (filename)
            realVariants = filter (None, self.properties ['VMAKEVARIANT'])
            result = name + '-' + string.join (realVariants,  '-') + ext
        else:
            result = filename
        return result

class VariantSet:
    def __init__ (self, list):
        # data is list of Configuration
        self.list = list

    def dump (self, comment = ''):
        trace ('=== VariantSet %s', comment)
        for entry in self.list:
            trace ('   %s', entry)

    def __getitem__ (self, index):
        return self.list [index]

    def getAllNames (self):
        set = {}
        for entry in self.list:
            for name in entry.variantNames:
                set [name] = 1
        result = set.keys ()
        result.sort ()
        return result

    def filter (self, variantName, returnList = false):
        list = []
        for entry in self.list:
            #if not entry.options:
            #    # empty option lists are ignored
            #    continue
            if variantName in entry.variantNames:
                list.append (entry)
        if returnList:
            result = list
        else:
            result = VariantSet (list)
        return result

    #
    # extract common option for one variant
    #
    def getAvailableOptions (self):
        result = []
        for entry in self.list:
            for optdef in entry.options:
                if optdef not in result:
                    result.append (optdef)
        return result

    def isCommonOption (self, optdef):
        for entry in self.list:
            if optdef not in entry.options:
                return false
        return true

    def getCommonOptions (self):
        options = self.getAvailableOptions ()
        result = filter (self.isCommonOption,  options)
        return result

    def removeOption (self, optdef):
        for entry in self.list:
            entry.options.remove (optdef)

    def removeOptions (self, options):
        for optdef in options:
            self.removeOption (optdef)

    #
    # extract common options for all variants
    #
    def extractCommonOptions (self, *allNames):
        result = []
        if not allNames:
            allNames = self.getAllNames ()
        for name in allNames:
            subset = self.filter (name)
            commonOptions = subset.getCommonOptions ()
            if commonOptions:
                newConfig = Configuration ((name,))
                newConfig.options = commonOptions
                result.append (newConfig)
            subset.removeOptions (commonOptions)
        return result

class LayerDict (UserDict):
    def addEntry (self, path, basename, entry, dspFile):
        try:
            list = self [path]
        except KeyError:
            list = []
            self[path] = list
        list.append ((basename, entry, dspFile))

def sortedItems (dict):
    """return a list of key-value tuples, sorted by key
    """
    items = dict.items ()
    items.sort ()
    return items

class OptionSampler:
    def __init__ (self):
        self.dict = {}

    def addOption (self, tool, option, value):
        try:
            toolOptions = self.dict [tool]
        except:
            toolOptions = {}
            self.dict [tool] = toolOptions
        try:
            optionValues = toolOptions [option]
        except:
            optionValues = {}
            toolOptions [option] = optionValues
        if value:
            optionValues [value] = 1

    def dump (self):
        for tool, options in sortedItems (self.dict):
            trace (tool)
            for option, values in sortedItems (option):
                trace ('    %s', options)
                for value, dummy in sortedItems (values):
                    trace ('        %s', value)

class WorkspaceParser:
    def __init__ (self, filename):
        self.short2long = {}
        self.dependencies = {}
        matchList = []
        for kind in ['Project', 'Depname']:
            matchList.append ((getattr (self, 're' + kind), getattr (self, 'match' + kind)))
        self.matcher = RegexEvents (matchList)
        self.basedir = os.path.dirname (os.path.abspath (filename))
        self.parseFile (filename)

    def parseFile (self, filename):
        self.matcher.parseFile (filename)

    reProject = re.compile ('Project: "([^"]+)"=([^ ]+) - Package Owner=<4>')
    def matchProject (self, match):
        short, long = match.group (1, 2)
        long = os.path.normpath (os.path.join (self.basedir, long))
        self.short2long [short] = long
        self.currentDependencies = []
        self.dependencies [long] = self.currentDependencies

    reDepname = re.compile (r'\s*Project_Dep_Name (.*)')
    def matchDepname (self, match):
        self.currentDependencies.append (match.group (1))

    def getDependencies (self):
        result = []
        items = self.dependencies.items ()
        items.sort ()
        for dsp, dependencies in items:
            absDep = map (self.transShort, dependencies)
            result.append ((dsp, absDep))
        return result

    def transShort (self, name):
        return self.short2long [name]

def reExpand (template, rex, match):
    variables = {}
    for varname, index in rex.groupindex.items ():
        variables [varname] = match.group (index)
    return template % variables

def unquote (str):
    if str[0] == '"' and str [-1] == '"':
        str = str [1:-1]
    return str

def split3 (filename):
    """split a filename into directory, basename, extension
    """
    directory, basename = os.path.split (filename)
    basename, extension = os.path.splitext (basename)
    return directory, basename, extension

def equalFiles (file1, file2):
    file1 = string.replace (file1, '\\', '/')
    return string.lower (file1) == string.lower (file2)

def setSplit (set1, set2):
    """split two lists into

       * elements only in set1
       * elements in both lists
       * elements only in set2
    """
    common = []
    rem1 = []
    rem2 = []
    for elem in set1:
        if elem in set2:
            common.append (elem)
        else:
            rem1.append (elem)
    for elem in set2:
        if elem in set1:
            pass
        else:
            rem2.append (elem)
    return rem1, common, rem2

def sortedDict (dict):
    result = dict.items ()
    result.sort ()
    return result

knownLibs = {
    'allapi'    : '::aps/api/api/AllApi.lib',
    'stpapi'    : '::aps/api/StpApi/StpApi.lib',
    'tloapi'    : '::aps/api/TloApi/TloApi.lib',
    'apsbase'   : '::aps/base/apsbase.lib',
    'apscore'   : '::aps/core/apscore.lib',
    'atpapi'    : '::atp/api/AtpApi/AtpApi.lib',
    'atpcoreapi': '::atp/api/AtpCoreApi/AtpCoreApi.lib',
    'atplib'    : '::atp/atplib/atplib.lib',
    'deltalib2' : '::atp/deltalib2/deltalib2.lib',
    'ltt'       : '::ltt/LTT.lib',
    'ts'        : '::TimeSeries/Ts.lib',
    'tsapi'     : '::TimeSeries/api/tsapi.lib',
    'tssnp'     : '::TimeSeries/TsSnp/TsSnp.lib',
}

def loadCache (filename):
    if not os.path.exists (filename):
        return knownLibs
    execfile (filename, globals ())
    return knownLibs

def writeCache (filename, knownLibs):
    stream = open (filename, 'wt')
    stream.write ('\nknownLibs = {\n')
    items = knownLibs.items ()
    items.sort ()
    for name, value in items:
        stream.write ('    %-20s: %s,\n' % (repr (name), repr (value)))
    stream.write ('}\n')
    stream.close ()

def readConfig (options):
    if os.path.exists (options.config):
        locals = {}
        execfile (options.config, locals)
        for varname in ['inDEV', 'outSys']:
            configValue = locals.get (varname)
            if configValue and not getattr (options, varname):
                setattr (options, varname, configValue)

def checkOptions (options):
    readConfig (options)
    if options.ignore:
        if not os.path.exists (options.ignore):
            sys.stderr.write ('dsp2vmake: Cannot open file "%s" (from -ignore)\n')
            sys.exit (2)
        lines = open (options.ignore, 'rt').readlines ()
        lines = map (string.strip, lines)
        lines = map (string.lower, lines)
        lines = filter (None, lines)
        options.ignorePatterns = lines
    else:
        options.ignorePatterns = []
    try:
        options.ignorePatterns.extend (map (string.lower, specialCases.filesToIgnore))
    except AttributeError:
        pass
    options.globals.append ('*.cpp -except')
    if not options.inDEV:
        fatalError ('options -inDEV not set')
    if not options.outSys:
        fatalError ('options -outSys not set')
    options.nowarn = not options.warn

def main (options, args):
    """translates MS Dev projects to vmake descriptions

    """
    checkOptions (options)
    if options.unifiedSys:
        outDesc = os.path.join (options.outSys, 'src')
    else:
        outDesc = os.path.join (options.outSys, 'desc')
    # knownLibs = loadCache (options.cache)
    translator = Dsp2Vmake (options.inDEV, outDesc, None, options)
    translator.getTargetKinds (args)
    if args:
        for arg in args:
            ext = os.path.splitext (arg) [1]
            if ext == '.dsw':
                translator.translateWorkspace (arg)
            else:
                translator.translateOneFile (arg)
    else:
        translator.translate (options.inDEV)
    translator.writeComDescriptions ()
    # writeCache (options.cache, knownLibs)

def _options ():
    configFile = os.path.join (os.environ ['TOOL'], 'data', 'dsp2vmake.conf')
    default_inDEV = os.environ.get ('SRC')
    try:
        default_outSys = os.path.join (os.environ ['OWN'], 'sys')
    except KeyError:
        default_outSys = None
    return [
        # (optstring, varname, typechar, default, help)
        ('inDEV', None, ':', default_inDEV, 'input directory'),
        ('outSys', None, ':', default_outSys, 'output directory'),
        ('config', None, ':', configFile, 'name of configFile'),
        ('unifiedSys', None, '', true, '$DESC == $SRC'),
        ('v', 'verbose', '', None, 'verbose output'),
        ('ignore', None, ':', None, """file containing file names to ignore.
             Matches anywhere in the file name
                 \\aps\\ would exclude all dirs under aps
                 xyz.dsp would exclude any file named xyz.dsp
             """),
        ('globals', None, '::', None, 'to insert in every .com'),
        #('cache', None, ':', 'dsp.cache', 'cache for dsp types'),
        ('debug', None, '', None, 'debug output'),
        ('nodep', None, '', None, 'don\'t generate dependency information'),
        ('warn', None, '', None, 'generate warnings'),
        ('ignoreVariant', None, '::', None, 'don\'t generate this variant'),
        ('onlyVariant', None, '::', None, 'generate only these variants'),
        ('dumbVariants', None, '', 1, 'don\'t split DebugSim into Debug, Sim'),
        ('smartVariants', None, '', None, 'split DebugSim into Debug, Sim'),
        ('version', None, '', "0.4", 'version information'),
        ]

if __name__ == '__main__':
    import optlib
    optlib.optMain2 (main, _options (), optvar = 'DSP2VMAKE_OPTION')

